/*
 * HFS (aka MAC) filesystem/handler for MorphOS/AmigaOS
 * Author: Marek 'March' Szyprowski <marek@amiga.pl>
 */

#include <exec/types.h>
#include <exec/execbase.h>
#include <exec/memory.h>
#include <dos/dosextens.h>
#include <dos/filehandler.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/utility.h>

#include <string.h>

#include "mac_fs.h"
#include "mac_protos.h"

        
LONG LockObject(struct FSSuper *sb, ULONG cid, ULONG parid, LONG axs, BPTR *res)
{
	LONG err = 0;
	struct ExtFileLock *fl;
	kprintf("\tLock Object cid %ld, parent id %ld\n", cid, parid);

	if ((fl = FS_AllocMem(sizeof(struct ExtFileLock))))
	{
		memset(fl, 0, sizeof(*fl));
		kprintf("\tCreating lock...\n");
		fl->fl_Key = cid;
		fl->fl_Access = axs;
		fl->fl_Task = glob->ourport;
		fl->fl_Volume = MKBADDR(sb->doslist);
		fl->fl_Link = sb->doslist->dol_misc.dol_volume.dol_LockList;
		fl->magic = ID_HFS_DISK;
		fl->sb = sb;
		fl->parid = parid;

		sb->doslist->dol_misc.dol_volume.dol_LockList = MKBADDR(fl);

		*res = MKBADDR(fl);
		kprintf("\tCreated lock %08lx\n", *res);
	}
	else
		err = ERROR_NO_FREE_STORE;
	return err;
}

LONG LockParent(struct FSSuper *sb, struct ExtFileLock *fl, LONG axs, BPTR *res)
{
	ULONG cid = fl->parid;
	LONG err;
	ULONG parent;

	if ((err = GetParentDir(sb, cid, &parent)) == 0)
		err = LockObject(sb, cid, parent, axs, res);
	return err;
}

LONG IsLockable(struct FSSuper *sb, ULONG cid, LONG mode)
{
	struct ExtFileLock *cur;

	for (cur=BADDR(sb->doslist->dol_misc.dol_volume.dol_LockList); cur; cur=BADDR(cur->fl_Link))
		if (cur->fl_Key == cid)
			if (cur->fl_Access == EXCLUSIVE_LOCK || mode == EXCLUSIVE_LOCK)
				return ERROR_OBJECT_IN_USE;
	return 0;
}

LONG FreeLockSB(struct FSSuper *sb, struct ExtFileLock *fl)
{
	LONG found = FALSE;

	if (sb == NULL)
		return ERROR_OBJECT_NOT_FOUND;

	if (fl == BADDR(sb->doslist->dol_misc.dol_volume.dol_LockList))
	{
		sb->doslist->dol_misc.dol_volume.dol_LockList = fl->fl_Link;
		found = TRUE;
	}
    else
	{
		struct ExtFileLock *prev = NULL, *ptr = BADDR(sb->doslist->dol_misc.dol_volume.dol_LockList);
		while (ptr != NULL)
		{
			if (ptr == fl)
			{
				prev->fl_Link = fl->fl_Link;
				found = TRUE;
				break;
			}
			prev = ptr;
			ptr = BADDR(ptr->fl_Link);
		}
	}
	if (found)
	{
		kprintf("\tFreeing lock.\n");
		fl->fl_Task = NULL;
		if (fl->dir)
			hfs_closedir(fl->dir);
		if (fl->file)
			hfs_close(fl->file);
		FS_FreeMem(fl);
		return 0;
	}
	return ERROR_OBJECT_NOT_FOUND;
}

LONG FreeLock(struct ExtFileLock *fl)
{
	LONG err = 0;
	struct FSSuper *ptr = glob->sblist, *prev=NULL;

	if (fl->magic != ID_HFS_DISK)
		return ERROR_OBJECT_WRONG_TYPE;

	if (fl->sb == glob->sb)
		return FreeLockSB(glob->sb, fl);

	while (ptr != NULL)
	{
		if (fl->sb == ptr)
		{
			if ((err = FreeLockSB(ptr, fl)) != 0)
				return err;

			if ((err = FreeUnusedDisk(ptr, prev)) == ERROR_OBJECT_IN_USE)
				err = 0; // ignore
			return err;
		}
		prev = ptr;
		ptr = ptr->next;
	}
	return ERROR_OBJECT_NOT_FOUND;
}

void FreeLockStructures(struct FSSuper *sb)
{
	struct ExtFileLock *fl;

	kprintf("\tFreeing internal structures from unrealesed locks\n");
	fl = BADDR(sb->doslist->dol_misc.dol_volume.dol_LockList);
	while (fl != NULL)
	{
		if (fl->dir)
			hfs_closedir(fl->dir);
		if (fl->file)
			hfs_close(fl->file);
		fl->dir = NULL;
		fl->file = NULL;
		kprintf("\t\tLocked cid %ld released\n", fl->fl_Key);
		fl = BADDR(fl->fl_Link);
	}
}

LONG RestoreLockStructures(struct FSSuper *sb)
{
	LONG err = 0;
    struct ExtFileLock *fl;

	kprintf("\tRestoring internal structures in unrealesed locks\n");
	fl = BADDR(sb->doslist->dol_misc.dol_volume.dol_LockList);
	while (err == 0 && fl != NULL)
	{
		ULONG tmp;
		
		kprintf("\t\tRestoring cid %ld lock\n", fl->fl_Key);
		fl->sb = sb;

		if (FileOpen(sb, fl) == 0)
			FileSeek(sb, fl, fl->lastpos, OFFSET_BEGINNING, &tmp);
		fl = BADDR(fl->fl_Link);

	}
	return err;
}

